@@ -1399,6 +1399,7 @@ PyObject *obj_Vector_new(PyTypeObject *type,PyObject *args,PyObject *kwds) {
 
         auto ptr = py::check_obj(type->tp_alloc(type,0));
         auto &base = reinterpret_cast<wrapped_type<n_vector>*>(ptr)->alloc_base();
+        new(&base) n_vector(dimension);
         
         if(values) {
             sized_iter itr(*values,dimension);
@@ -1576,6 +1577,12 @@ PySequenceMethods obj_Matrix_sequence_methods = {
     NULL
 };
 
+PyObject *obj_Matrix_reflection(PyObject*,PyObject *arg) {
+    try {
+        return to_pyobject(n_matrix::reflection(from_pyobject<n_vector>(arg)));
+    } PY_EXCEPT_HANDLERS(NULL)
+}
+
 PyObject *obj_Matrix_scale(PyObject*,PyObject *args) {
     try {
         if(PyTuple_GET_SIZE(args) == 1) {
@@ -1643,6 +1650,7 @@ PyObject *obj_Matrix_transpose(wrapped_type<n_matrix> *self,PyObject*) {
 }
 
 PyMethodDef obj_Matrix_methods[] = {
+    {"reflection",reinterpret_cast<PyCFunction>(&obj_Matrix_reflection),METH_O|METH_STATIC,NULL},
     {"scale",reinterpret_cast<PyCFunction>(&obj_Matrix_scale),METH_VARARGS|METH_STATIC,NULL},
     {"rotation",reinterpret_cast<PyCFunction>(&obj_Matrix_rotation),METH_VARARGS|METH_KEYWORDS|METH_STATIC,NULL},
     {"identity",reinterpret_cast<PyCFunction>(&obj_Matrix_identity),METH_O|METH_STATIC,NULL},
